#! /usr/bin/env python3

import sys
try:
    from mpi4py import MPI
except RuntimeError:
    print("Error loading mpi4py. Make sure MPI is available and mpi4py is installed")
    print("You might have to load the mpi module on a cluster (module load mpi)")
    sys.exit(1)
except ModuleNotFoundError:
    print("mpi4py module not found.")
import numpy as np
from humanize import naturalsize, intword, intcomma
import h5py
import blosc2
from time import sleep, time
import logging
import argparse
import zipfile
from pathlib import Path
import json

from importlib.metadata import version

__version__ = version("hdf5vault")

from hdf5_archive_info import HDF5_ARCHIVE_INFO

# define MPI communication tags 
request_filename_tag = 1  # worker from master
send_filename_tag = 2     # master to worker
file_compressed_tag = 3   # worker to master
cdata_wo_to_co_tag = 4    # worker to write coordinator
cdata_co_to_wr_tag = 5    # write coordinator to writer
request_cdata_tag = 6     # writer to write coordinator
data_written_tag = 7      # writer to master
wait_tag = 97             # write coordinator to writer
compression_complete_tag = 98 # master to write coordinator
stop_tag = 99             # master to worker and write coordinator to writers

logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout)

# splits file into chunks (if larger than chunksize) and compresses each chunk
# does not compress chunk if no storage gains are found
# returns an iterator over chunk numbers; each item is a dict with compressed buffer, chunk number and compression method 
def compress_file(filepath: Path, method: str, comm:MPI.Comm, 
                  chunksize:int = 2_000_000_000, 
                  nthreads=4, clevel=8, use_dict=False):

    from math import ceil
    
    max_chunk_size = 2_147_483_600
    filesize=filepath.stat().st_size

    if chunksize > max_chunk_size:
        logger.error(f"chunksize {chunksize} is larger than maximum of {max_chunk_size}")
        comm.Abort()

    nchunks = max(ceil(filesize / chunksize), 1) # nchunks is 1 even for an empty file

    if method.lower() == "blosc2":

        cparams = { 
                "codec": blosc2.Codec.ZSTD,
                "typesize": 8,
                "clevel": clevel,
                "use_dict": use_dict,
                "nthreads": nthreads
                }

    else: 
        cparams = None

    if not filepath.is_file():
        logger.error(f"{filepath} does not exist or is not a regular file")
        comm.Abort()    

    with filepath.open('rb') as fid:
        for n in range(nchunks):
            t0=time()
            buffdata = fid.read(chunksize)
            t1=time()

            if method.lower() == 'none':
                cdata = buffdata
                used_method = None
                compression_time = 0.

            else:
                cdata=None
                t2=time()
    
                try:
                    cdata=blosc2.compress2(buffdata, **cparams)
            
                except RuntimeError:
                    if use_dict:
                        cparams_nodict = cparams.copy()
                        cparams_nodict["use_dict"] = False
                        logger.info(f"file {filepath}, chunk {n} compressed with use_dict=False")
                        cdata=blosc2.compress2(buffdata, **cparams_nodict)
                    else:
                        logger.error(f"file {filepath}, chunk {n} can not be compressed, even with dict=False.")
                        comm.Abort()
                        
                except Exception as e:
                    logger.error(f"Received error {e} while compressing {filepath} chunk {n}")
                    comm.Abort()
    
                assert cdata is not None, "compressed data is none. this should not happen"
                t3=time()
                compression_time = t3-t2
            
            read_time = t1-t0

            # if compression results in no storage saving, chunk is stored uncompressed
            if len(cdata) >= len(buffdata):
                cdata = buffdata
                used_method = None
            else:
                used_method = method

            if nchunks > 1:
                chunk_number = n
            else:
                chunk_number = None

            out = {"buffer": cdata, 
                   "chunk": chunk_number, 
                   "compression": used_method, 
                   "raw_size": len(buffdata),
                   "read_time": read_time,
                   "total_chunks": nchunks,
                   "compression_time": compression_time}

            yield out

def iter_files(tdir: Path):
    for p in tdir.rglob('*'):
        if p.is_file():
            yield p.relative_to(tdir).as_posix()

# determine wait time based on size that has already been written to file
# the scope is to even out HDF5 file sizes, making writers that have already
# written a lot of data wait longer
def determine_wait_time(wnum, filesizes, min_wait=0.001, max_wait=0.010):
    min_size = min(filesizes.values())
    max_size = max(filesizes.values())

    dw = max_wait - min_wait
    dsize = max_size - min_size
    if dsize == 0:
        return min_wait
    asize = filesizes[wnum]

    return min_wait + (asize - min_size) / dsize * dw

# HDF5Vault uses many processes to open files and compress their data, and (typically) 
# fewer processes to write that data into an archive.

# There are four different type of tasks a process can do, depending on its rank:

# master: obtain list of files, inform compressors which files to compress. track progress.
# compressors: obtain name of file to read and compress. send compressed data to write coordinator. repeat.
# write coordinator: receive compressed data from compressors. send off to writers upon request.
# writer: ask write coordinator which data to write. write it to file. repeat.

# Master (rank=0): obtains list of files and listens to incoming connections. 1 requests types:
# - RECV: give me the path to a new file that I can compress (worker)
# - SEND: here's a file to compress
# - RECV: I have compressed file X and sent it off to a writer
# Write coordinator (rank=1): Receives compressed data and sends it to writers
# - RECV: here's data for file X I compressed (from worker)
# - RECV: send me more data (from writer)
# - SEND: write this to disk (to worker)
# writer (rank=2 to 2+nwriters): 
# - RECV: write this data to disk
# worker (rank=1+nwriters to ncpus):
# - SEND: send me a filename (to master) 
# - RECV: filename (from master)
# - SEND: here's a compressed file (to write coordinator)
# - SEND: i compressed this file and sent it to the write coordinator (to master)

def master_function(tdir: Path, nworkers: int, comm: MPI.Intracomm, comm_writers: MPI.Intracomm, prog_step: int):
    filelist=[]

    if not tdir.is_dir():
        logger.error(f"{tdir} is not a directory")
        comm.Abort(1)

    t0=time()
    file_iter = iter_files(tdir)
    t1=time()
    list_files_time=t1-t0

    # dict with key filename and value number of chunks in that file
    files_compressed = {}

    # a dict with an entry for each file written to disk. each entry is a dict with chunk_number: writer_rank
    files_written = {}

    workers_finished = set()

    status = MPI.Status()

    nchunks_written = 0
    chunk_counter=0

    file_counter=0

    # loop continues until all workers have been sent stop tag and all files have been confirmed stored
    while len(workers_finished) < nworkers or nchunks_written < chunk_counter:
        logger.debug(f'MA: waiting for connections')
        rdata = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
        worker_rank = status.Get_source()
        logger.debug(f'MA: received message with tag {status.Get_tag()} from rank {worker_rank}')

        if status.Get_tag() == request_filename_tag:
            filename = next(file_iter, False) # return False if iterator is exhausted
            if filename: 
                files_written[filename] = []
                comm.send(filename, dest=worker_rank, tag=send_filename_tag)
                logger.debug(f'MA: sending {filename} to {worker_rank}')
                if (file_counter % 1000 == 0):
                    logger.info(f'Found {file_counter} files ...')
                file_counter+=1
            else:
                comm.send(None, dest=worker_rank, tag=stop_tag)
                logger.debug(f'MA: sending stop tag to {worker_rank}')
                workers_finished.add(worker_rank)

        elif status.Get_tag() == file_compressed_tag:
            compressed_file=rdata["filename"]
            files_compressed[compressed_file] = rdata["total_chunks"]
            logger.debug(f'MA: received compression confirmation from {worker_rank} for file {compressed_file}')

            chunk_counter+=rdata["total_chunks"]

        elif status.Get_tag() == data_written_tag:
            written_file=rdata["filename"]
            chunk=rdata["chunk"]
            # continue here: this will be list, with each entry a dict
            #files_written[written_file][chunk] = worker_rank
            files_written[written_file].append(
                {"chunk": chunk, "archive_number": rdata["archive_number"], "dataset": rdata["dataset"], "compression": rdata["compression"] } 
            )
            
            nchunks_written += 1
            logger.debug(f'MA: received write confirmation from {worker_rank}')

            if nchunks_written % prog_step == 0:
                logger.info(f'{nchunks_written}  files or chunks written to archive.')

    logger.debug(f'MA: sending stop tag to rank=1 (write coordinator)')
    comm.send(None, dest=1, tag=compression_complete_tag)

    logger.debug(f'MA: asserting all chunks have been written')

    success = True
    for key in files_written.keys():
        #if files_written[key] is False:
        if len(files_written[key]) != files_compressed[key]:
            logger.error(f"MA: file {key} has not been confirmed as written (completely). Number of chunks = {files_compressed[key]}, chunks_written = {len(files_written[key])}.")
            success = False

    # exchange archive info with writers
    logger.debug('MA: broadcasting archive info ')
    comm_writers.bcast(files_written, root=0)

    if not success:
        comm.Abort(1)
    else:
        logger.info('MA: all chunks confirmed as written')

    return file_counter, chunk_counter, list_files_time

def write_coordinator_function(nwriters: int, comm: MPI.Intracomm):
    status = MPI.Status()

    compression_completed = False
    writers_stopped = set()

    cdata = []
    
    filesizes = {n: 0 for n in range(nwriters)}

    # the write coordinator is active until
    # - all files have been compressed and sent off to write coordinator) AND
    # - all writers have been sent a stop_tag by the write coordinator

    while not compression_completed or len(writers_stopped) < nwriters:
        logger.debug(f"CO: waiting for messages")
        ndata=comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)

        msg_source = status.Get_source()

        # a compressor (worker) is delivering data that has been compressed
        if status.Get_tag() == cdata_wo_to_co_tag:
            logger.debug(f'CO: received data of type {type(ndata)} from {msg_source}')
            """
            filename=ndata["filename"]
            buffer_cmp=ndata["buffer_cmp"]

            if filename in cdata.keys():
                logger.error(f'CO: data for file {filename} is already in cdata')
                comm.Abort()

            cdata[filename] = buffer_cmp
            """
            # in this version, the fifo stack is just a list, and I add the whole dict
            # as received to it
            cdata.append(ndata)

        # a writer is requesting data to write
        elif status.Get_tag() == request_cdata_tag:
            logger.debug(f'CO: received data request from {msg_source}')
            if msg_source == 0 or msg_source >= nwriters + 2:
                logger.error("CO: received cdata request from unexpected source")
                comm.Abort(1)

            writer_rank = msg_source - 2

            # 3 possibilities: 
            # a. there is data to write -> send it to the writer, and remove it from buffer
            # b. there is currently no data to write available, but compression ongoing. send wait_tag to writer 
            # c. there is no data, and compression is complete. send stop_tag to writer

            if len(cdata) == 0:
                if not compression_completed:
                    sleeptime = determine_wait_time(writer_rank, filesizes)
                    comm.send(sleeptime, dest=msg_source, tag=wait_tag)
                    logger.debug(f'CO: sent sleep tag (sleeptime = {sleeptime}) to {msg_source}')
                else:
                    comm.send(None, dest=msg_source, tag=stop_tag)
                    writers_stopped.add(msg_source)
                    logger.debug(f"CO: sent stop tag to writer {msg_source}")
            else:
                #filename = list(cdata.keys())[0]
                #ndata = cdata.pop(filename)
                ndata=cdata.pop()

                comm.send(#{"filename": filename,
                          #  "buffer_cmp": ndata}, 
                          ndata,
                            dest=msg_source, 
                            tag=cdata_co_to_wr_tag)

                filesizes[writer_rank] += len(ndata["buffer"])

        elif status.Get_tag() == compression_complete_tag and status.Get_source() == 0:
            logger.debug(f'CO: received compression complete tag from {msg_source}')
            compression_completed=True

    return

def writer_function(nwriters: int, comm: MPI.Intracomm, comm_writers: MPI.Intracomm, rank: int, 
                    archive_basename: str, cmethod: str, 
                    chunksize: int, 
                    usezip: bool=False):

    status = MPI.Status()
    waittime = 0 # time writers spend waiting
    metatime = 0 # time for writing metadata
    writetime = 0 # time for writing bytes
    wrcotime = 0 # time for communication (writers)

    stopped=False

    # each writers writes to its own file
    if nwriters == 1:
        multifile = False
        filenum = 0
    else:
        multifile = True
        filenum = rank - 2

    ext = "h5" if not usezip else "zip"

    if multifile:
        ndecimals = len(str(nwriters))
        h5filename=f"{archive_basename}_{filenum:0{ndecimals}d}.{ext}"
        print(h5filename)
    else:
        h5filename=f"{archive_basename}.{ext}"
    
    all_dsets = []

    if not usezip: 
        ctx=h5py.File(h5filename, 'w')

    else:
        ctx=zipfile.ZipFile(h5filename, mode='w', compression=zipfile.ZIP_STORED)

    with ctx as archive:
        while not stopped:
            if not usezip:
                archive.attrs["__Description__"] = HDF5_ARCHIVE_INFO
                archive.attrs["__HDF5Vault_version__"] = __version__
                if multifile:
                    archive.attrs["__HDF5Vault_archive_file_number__"] = filenum
                    archive.attrs["__HDF5Vault_number_of_archive_files__"] = nwriters
                archive.attrs["__HDF5Vault_chunksize__"] = chunksize

            logger.debug(f'WR {rank} is waiting for data to write')
            comm.send(None, dest=1, tag=request_cdata_tag)

            ta=time()
            ndata=comm.recv(source=1, tag=MPI.ANY_TAG, status=status)

            msg_tag = status.Get_tag()

            if msg_tag == wait_tag:
                logger.debug(f'WR {rank} received wait tag')
                sleep(ndata) # the transmitted number is the wait time
                waittime += ndata

            elif msg_tag == stop_tag:
                logger.debug(f'WR {rank} received stop tag')
                stopped=True

            elif msg_tag == cdata_co_to_wr_tag:
                if ndata is None:
                    logger.error("WR: received None data to write")
                    comm.Abort(1)
                logger.debug(f"WR {rank} received data to write for file {ndata['filename']}")

                # now dump the data to disk
                t0 = time()
                #if cmethod.lower() == 'none':

                if ndata["chunk"] is None:
                    dset_basename = ndata["filename"]
                else:
                    chunknumber=ndata["chunk"]
                    total_chunks = ndata["total_chunks"]
                    ndecimals = len(str(total_chunks))
                    dset_basename = ndata["filename"] + f"____part{chunknumber:0{ndecimals}d}"

                if ndata['compression'] is None:
                    dset_name = dset_basename
                else:
                    dset_name = dset_basename + "." + cmethod

                if not usezip:
                    dset_shape = len(ndata["buffer"])
                    all_dsets.append(ndata["filename"])

                    archive.create_dataset(dset_name, 
                                        shape=dset_shape,
                                        dtype=np.byte,
                                        compression=None)

                t1=time()

                if not usezip:
                    archive[dset_name][:] = np.frombuffer(ndata["buffer"], dtype=np.byte)

                    if ndata["chunk"] is not None:
                        archive[dset_name].attrs['chunk'] = chunknumber
                        archive[dset_name].attrs['number_of_chunks'] = total_chunks

                    # compression is now file-specific
                    archive[dset_name].attrs['compression'] = ndata["compression"] if ndata["compression"] is not None else "None"
                    archive.flush()


                else:
                    archive.writestr(dset_name, ndata["buffer"])

                t2=time()

                wrcotime += t0 - ta
                metatime += t1 - t0
                writetime += t2 - t1

                logger.debug(f"WR: sending write confirmation from rank {rank} to rank 0")

                # confirmation data to be sent back to master
                conf_data={"filename": ndata["filename"],
                           "chunk": ndata["chunk"],
                           "dataset": dset_name,
                           "archive_number": filenum,
                           "compression": ndata["compression"]}

                comm.send(conf_data, 
                          dest=0, tag=data_written_tag)

        # end of while loop

        logger.debug('WR: waiting for archive info broadcast')
        archive_info={}
        archive_info = comm_writers.bcast(archive_info, root=0 )
        # create HDF5 dataset with list of files
        logger.debug(f'Writing archive info attribute to {h5filename}.') # formerly __filelist__
        if not usezip:
            archive.attrs["__HDF5Vault_info__"] = json.dumps(archive_info, indent=3)

        else:
            archive.writestr("__HDF5Vault_info__", json.dumps(archive_info, indent=3))


    # end of with, closing file

    return waittime, metatime, writetime, wrcotime

def compressor_function(comm: MPI.Intracomm, rank: int, tdir: Path, cmethod: str, nthreads: int, clevel: int, use_dict: bool, chunksize:int):
    status = MPI.Status()

    finished = False

    # keep track of times and sizes

    readtime = 0 # time for reading
    comptime = 0 # time for compression
    raw_size = 0 # raw data size
    comp_size = 0 # compressed data size
    wocotime = 0 # time for communication (compressors)

    while not finished:
        logger.debug(f"WO: rank {rank}: Requesting new file from master")
        comm.send(None, dest=0, tag=request_filename_tag)
        filename=comm.recv(source=0, tag=MPI.ANY_TAG, status=status)

        if status.Get_tag() == stop_tag:
            finished = True
            logger.debug(f"WO: finished on worker {rank}")
        else:

            # do compression
            filepath = tdir.joinpath(filename)
            logger.debug(f"WO: opening {filepath} on rank {rank}")

            #buffer_cmp=compress_buffer(buffer, cmethod, nthreads=nthreads, clevel=clevel, use_dict=use_dict, filename=filename)
            for cfiled in compress_file(filepath, cmethod, comm, chunksize=chunksize, nthreads=nthreads, clevel=clevel, use_dict=use_dict):
                if cfiled["buffer"] is None:
                    comm.Abort(1)

                raw_size += cfiled["raw_size"]
                comp_size += len(cfiled["buffer"])

                logger.debug(f"WO: sending compressed data from {filename} (chunk {cfiled['chunk']}) from rank {rank} to rank 1")

                # add filename 
                cfiled["filename"] = filename
                t2=time()
                comm.send(cfiled,
                        dest=1, # to write coordinator
                        tag=cdata_wo_to_co_tag)

                t3=time()

                readtime += cfiled["read_time"]
                comptime += cfiled["compression_time"]
                wocotime += t3-t2

                # try removing this to speed things up
                #req.wait()

            logger.debug(f"WO: sending compression confirmation from rank {rank} to rank 0")
            # sending to rank 0 the filename and the number of chunks that file was split into
            comm.send({key: cfiled[key] for key in ["filename","total_chunks"]},
                       dest=0, tag=file_compressed_tag)
    
    return readtime, comptime, wocotime, comp_size, raw_size

def create_hdf5_archive(tdir,
                        h5file_basename,
                        nwriters=2, 
                        cmethod='blosc2', 
                        nthreads=4,
                        clevel=8,
                        use_dict=True,
                        prog_step=1000,
                        usezip=False,
                        chunksize=100_000_000):

    comm=MPI.COMM_WORLD

    ncpus=comm.Get_size()
    rank=comm.Get_rank()

    if rank==0 and usezip:
        print("Using zip format.")

    assert ncpus > (2 + nwriters), 'The number of ranks must be at least 2 + nwriters + 1'

    nworkers = ncpus - 2 - nwriters

    """
    Example:
    0 master
    1 write coordinator
    2 writer
    3 writer
    4 worker
    """

    # create a communicator with the master and all the workers
    if rank == 0 or rank >= 2+nwriters:
        color = 0 # master
        readtime = 0
        comptime = 0
        raw_size = 0
        comp_size = 0
        wocotime = 0
    else:
        color = 1
    comm_workers=comm.Split(color, 0)
        
    # create a communicator with the master and all the writers
    if rank == 0 or rank > 1 and rank < 2+nwriters:
        color2 = 0
        metatime = 0
        writetime = 0
        wrcotime = 0
        waittime = 0
    else:
        color2 = 1
    comm_writers=comm.Split(color2, 0)

    status = MPI.Status()

    time_begin = time()

    # master
    if rank == 0:
        nfiles,nchunks,list_files_time=master_function(tdir, nworkers, comm, comm_writers,prog_step)

    # write coordinator
    elif rank==1:
        write_coordinator_function(nwriters, comm)
        
    # writer
    elif rank > 1 and rank < 2+nwriters:
        waittime,metatime,writetime,wrcotime=writer_function(nwriters, comm, comm_writers, rank, h5file_basename, cmethod, 
                                                             chunksize, usezip)

    # compressor
    elif rank >= nwriters+2:
        readtime,comptime,wocotime,comp_size,raw_size=compressor_function(comm, rank, tdir, cmethod, nthreads, 
                                                                          clevel, use_dict, chunksize)
        
    logger.debug(f'ALL: rank {rank} at end of code.')
    comm.Barrier()
    time_end = time()

    # collecting statistics

    if color == 0:
        readtime = comm_workers.allreduce(readtime, op=MPI.SUM)
        comptime = comm_workers.allreduce(comptime, op=MPI.SUM)
        raw_size = comm_workers.allreduce(raw_size, op=MPI.SUM)
        comp_size = comm_workers.allreduce(comp_size, op=MPI.SUM)
        wocotime = comm_workers.allreduce(wocotime, op=MPI.SUM)

    if color2 == 0:
        metatime = comm_writers.allreduce(metatime, op=MPI.SUM)
        writetime = comm_writers.allreduce(writetime, op=MPI.SUM)
        wrcotime = comm_writers.allreduce(wrcotime, op=MPI.SUM)
        waittime = comm_writers.allreduce(waittime, op=MPI.SUM)

    if rank==0:
        sratio = raw_size / comp_size
        tp = naturalsize(comp_size / (writetime / nwriters))

        #effective throughput, data written during session divided by wallclock time
        tpe = naturalsize(comp_size / (time_end-time_begin))

        print(f'------------- Compression and performance summary ----------')
        print(f'Raw data size: {naturalsize(raw_size)}, {intword(nfiles)} files ({intcomma(raw_size)} bytes, {intcomma(nfiles)} files, {intcomma(nchunks)} chunks)')
        print(f'Compressed data size {cmethod}: {naturalsize(comp_size)} ({comp_size} bytes); storage ratio = {sratio}')
        print(f'Wallclock time to read, compress and write ({nworkers} compressors ({nthreads} threads), {nwriters} writers: {(time_end-time_begin):.3f} seconds.')
        print(f'Wallclock time retrieve list of files to archive (serial): {list_files_time:.3f} seconds.')
        print(f'Total CPU time spent on reading : {readtime:.3f} seconds.')
        print(f'Total CPU time spent on compressing ({cmethod}): {comptime:.3f} seconds.')
        print(f'Total CPU time spent on writing : {writetime:.3f} seconds, throughput {tp}/s (effective {tpe}/s).')
        print(f'Total CPU time spent on metadata : {metatime:.3f} seconds.')
        print(f'Total CPU time spent on communication by workers: {wocotime:.3f} seconds.')
        print(f'Total CPU time spent on communication by writer(s): {wrcotime:.3f} seconds.')
        print(f'Total CPU time spent idling by writer(s): {waittime:.3f} seconds.')

def main():
    parser = argparse.ArgumentParser(
                    prog='hdf5vault_create',
                    description=f'MPI tool for parallel archiving of large number of files in HDF5 container (version {__version__})')

    parser.add_argument('target_directory', type=Path, help='name of directory to archive')
    parser.add_argument('HDF5_archive_file_base', help='base of HDF5 archive file name')
    parser.add_argument('-w', '--writers', type=int, default=2, help='number of writers and archive files (default: 2)')
    parser.add_argument('-t', '--threads', type=int, default=4, help='number of threads per compressor (default: 4)')
    parser.add_argument('-c', '--clevel', type=int, default=8, help='compression level (1-9, default: 8)')
    parser.add_argument('-m', '--cmethod', type=str, default='blosc2', help='compression method: blosc2 (default) or None)')
    parser.add_argument('-d', '--use_dict', action='store_true', help='use dictionaries in compression (default: False)')
    parser.add_argument('-z', '--zip', action='store_true', help='create ZIP container(s) instead of HDF5 (default: False)')
    parser.add_argument('-s', '--chunksize', type=int, default=1_000_000_000, help='maximum chunk size to compress, in bytes (default 1 GB, max 2.1 GB)')
    parser.add_argument('-e', '--debug', action="store_true", help='print debugging information')
    parser.add_argument('-q', '--quiet', action="store_true", help='minimal output')

    args=parser.parse_args()

    if args.debug:
        logger.setLevel(logging.DEBUG)
    elif args.quiet:
        logger.setLevel(logging.ERROR)
    else:
        logger.setLevel(logging.INFO)

    assert args.clevel > 0 and args.clevel < 9, "clevel must be between 1 and 9."

    create_hdf5_archive(args.target_directory,
                        args.HDF5_archive_file_base,
                        nwriters=args.writers,
                        clevel=args.clevel,
                        use_dict=args.use_dict,
                        nthreads=args.threads,
                        cmethod=args.cmethod,
                        usezip=args.zip,
                        chunksize=args.chunksize
                       )
                        

    return

if __name__ == "__main__":
    main()



import h5py
import logging
from pathlib import Path
import json
import zipfile

# determine archive format (h5 or zip) from multiple files
def determine_format(files: list[Path]):
    ext = set([Path(a).suffix.lstrip('.') for a in files])
    assert len(ext) == 1, 'all archives must be of same type (have same extension)'

    cext=ext.pop().lower()

    if cext not in ["h5", "zip"]: 
        return False

    return cext

def check_archive_files(archive_files: list, logger: logging.RootLogger):
    n_archives = len(archive_files)

    multifile = False if n_archives == 1 else True

    filelist_inarch = dict()

    numbered_archives={}

    qc_passed=True

    for n,archive_file in enumerate(archive_files):
        logger.info(f'Reading HDF5 attributes from {archive_file}...')
        if not Path(archive_file).is_file():
            logger.error(f"archive file {archive_file} does not exist or is not a regular file.")
            return None, None, False

        with h5py.File(archive_file, 'r') as hfile:
            if "__HDF5Vault_version__" not in hfile.attrs.keys():
                logger.error("This archive was created with a prior version < 0.3.0 of HDFVault and can not be verified with this version.")
                return None, None, False
            if multifile:
                n_archives2=hfile.attrs["__HDF5Vault_number_of_archive_files__"]
                if n_archives != n_archives2:
                    logger.error(f"number of archives in {archive_file} is inconsistent with number of files passed to program.")
                    qc_passed=False

                filenum=hfile.attrs["__HDF5Vault_archive_file_number__"]
            else:
                if "__HDF5Vault_number_of_archive_files__" in hfile.attrs.keys():
                    logger.error(f"{archive_file} attribute suggets multi-file-archive, but only one HdF5 file was provided.")
                    qc_passed=False
                filenum=0

            numbered_archives[int(filenum)] = archive_file

            # filelist_inarch is a dictionary where each key is a file name
            if n==0:
                filelist_inarch=json.loads(hfile.attrs["__HDF5Vault_info__"])
            else:
                if "__HDF5Vault_info__" not in hfile.attrs.keys():
                    logger.warning(f"Error. Archive {archive_file} does not contain entry __HDF5Vault_info__.")
            
    if set(numbered_archives.keys()) != set(range(n_archives)):
        logger.error('inconsistent file numbering')
        qc_passed=False

    return filelist_inarch, numbered_archives, qc_passed


def check_zip_files(archive_files: list[str], logger: logging.RootLogger):
    archive_file = archive_files[0]

    if not Path(archive_file).is_file():
        return None, False

    ctx=zipfile.ZipFile(archive_file, 'r')

    try:
        filelist_inarch=json.loads(ctx.read('__HDF5Vault_info__'))

    except KeyError:
        logger.error(f'Could not read __HDF5Vault__ from {archive_file}.  Zipfile was not created by HDF5Vault >= 0.3.0 or is incomplete.')
        return None, False

    except:
        logger.error(f'Could not parse __HDF5Vault__ from {archive_file}.')
        return None, False

    numbered_archives={}

    for n,archive_file in enumerate(archive_files):
        archnum=int(Path(archive_file).stem.split("_")[-1])
        if archnum != n:
            logger.warning(f"Inferred archive number {archnum} inconsistent with order as passed {n}. Aborting.")
        numbered_archives[n] = archive_file
    
    return filelist_inarch, numbered_archives, True


#! /usr/bin/env python

# Carries out comparison of checksum in files from given directory
# and in data from HDF5 archive.

GET_FILENAME_TAG = 0
REQUEST_FILENAME_TAG = 1

import argparse
from mpi4py import MPI
from pathlib import Path
from sys import argv
#from os.path import join, isdir, isfile, getsize, basename, dirname, relpath
import numpy as np
import h5py
from sys import stdout
import time
import blosc2
import hashlib
import pandas as pd
from functools import reduce
import logging
from os import walk
from humanize import intcomma
import json
import zipfile

from common import determine_format, check_archive_files, check_zip_files

logger = logging.getLogger(__name__)
logging.basicConfig(stream=stdout)

def get_files_to_check(tdir: str) -> list:
    filelist=[]

    for root,dirs,files in walk(tdir):
        for file in files:
            filelist.append(join(root, file))
        print(f"\rfound {intcomma(len(filelist))} files", end="")

    filelist=[relpath(item, tdir) for item in filelist]

    return filelist

# flatten a lists of list (returned by mpi.gather) to list:
def flatten_list(inlist: list[list]) -> list:
    return reduce(lambda x,y: x+y, inlist)

def run_verification(archive_files: list, filelist_inarch: list, target_dir: Path, archive_type: str, rank: int, comm: MPI.Comm, progress_step:int=100):
    success = True
    error_files = []

    # checksum information to be collected and stored
    filesize_ondisk = []
    filesize_inarch = []

    md5_ondisk = []
    md5_inarch = []

    file_passed = []
    number_of_chunks = []

    filelist = []

    # open all HDF5 files
    if archive_type == 'h5':
        hfile={n: h5py.File(archive_files[n], 'r') for n in archive_files.keys()}
    else:
        hfile={n: zipfile.ZipFile(archive_files[n], 'r') for n in archive_files.keys()}

    finished = False
    n=0

    while not finished:
        if rank==1:
            if n % progress_step == 0:
                logger.info(f"Processing file {n} out of {len(filelist_inarch)} on rank {rank}")

        logger.debug(f'rank {rank}: requesting file from 0')
        comm.send(None, dest=0, tag=REQUEST_FILENAME_TAG)
        logger.debug(f'rank {rank}: waiting to receive file to verify')
        filename = comm.recv(source=0, tag=GET_FILENAME_TAG)

        if filename is None:
            logger.debug(f'rank {rank}: no more files left to verify')
            break

        diskfile = target_dir.joinpath(filename)

        filelist.append(filename)

        odata=open(diskfile, "rb").read()
        filesize_ondisk.append(len(odata))

        logger.debug(f"Rank {rank}: looking for file {filename} in archives ...")

        # find in which archive this dataset is stored

        if filename not in filelist_inarch.keys():
            success=False
            logger.error(f"FAIL: {filename} not found in any archive file")
            error_files.append(filename + "(missing)")
            filesize_inarch.append(0)
            md5_inarch.append('N/A')
            md5_ondisk.append('N/A')
            file_passed.append(False)
        else:
            # loop over all chunks ## CONTINUE HERE
            ainfo = filelist_inarch[filename]

            rdata = b''
            nchunks = len(ainfo)

            # in the file info dictionary, each file contains a list of dicts with chunknumber, archive, dataset name, and compression
            # this list may not be sorted ascending by chunk number
            # create dictionary that maps chunk number to entry in list
            cnmap = {a['chunk'] if a['chunk'] is not None else 0: p for p,a in enumerate(ainfo)}
            for chunknum in range(nchunks):
                cinfo = ainfo[cnmap[chunknum]]

                archnum = cinfo["archive_number"]
                dataset=cinfo["dataset"]
                if archive_type == 'h5':
                    compdata=hfile[archnum][dataset][:].tobytes()
                elif archive_type == 'zip':
                    compdata=hfile[archnum].read(dataset)

                if cinfo["compression"] is None:
                    rdata += compdata
                elif cinfo["compression"].lower() in ['blosc','blosc2']:
                    rdata += blosc2.decompress(compdata)
                else:
                    logger.error(f"Unknown compression method {cinfo['compression']} for file {filename}, part {chunknum}")
                    comm.Abort()
            
            number_of_chunks.append(nchunks)

            filesize_inarch.append(len(rdata))

            md5_ondisk.append(hashlib.md5(odata).hexdigest())
            md5_inarch.append(hashlib.md5(rdata).hexdigest())

            if len(odata) != len(rdata):
                success=False
                logger.error(f"FAIL: size of {diskfile} ({len(odata)}) does not match size of dataset {filename} ({len(rdata)}) in archive(s)")
                error_files.append(diskfile.as_posix() + "(size)")
                file_passed.append(False)
            else:

                if md5_ondisk[-1] != md5_inarch[-1]:
                    success=False
                    logger.error(f"FAIL: checksum {md5_ondisk[-1]} of {diskfile} != {md5_inarch[-1]} of {filename} in archive(s)")
                    error_files.append(diskfile.as_posix() + "(md5)")
                    file_passed.append(False)
                else:
                    file_passed.append(True)
        n+=1

    for f in hfile.values():
        f.close()

    return success, file_passed, error_files, filesize_ondisk, filesize_inarch, md5_ondisk, md5_inarch, filelist, number_of_chunks

# this function is called by rank 0.  It scans the target directory for files and continuously passed the
# names of files it found to the other ranks, which do the verification
def scan_and_distribute_files(target_dir:Path, comm: MPI.Comm, ncpus:int, progress_step:int=100):
    if not target_dir.is_dir():
        logger.error(f"Target directory {target_dir} not found. Aborting.")
        comm.Abort()

    status=MPI.Status()
    nfiles_ondisk=0

    filelist_ondisk=[]

    for p in target_dir.rglob('*'):
        if p.is_file():
            filename_r = p.relative_to(target_dir)
            logger.debug("Rank 0: accepting request for filename...")

            comm.recv(None, source=MPI.ANY_SOURCE, tag=REQUEST_FILENAME_TAG, status=status)

            destination=status.Get_source()

            logger.debug(f"Rank 0: ready to send out filename {filename_r} to {destination}")
            comm.send(filename_r.as_posix(), dest=destination, tag=GET_FILENAME_TAG)
            nfiles_ondisk+=1

            if nfiles_ondisk % progress_step == 0:
                logger.info(f"Found {nfiles_ondisk} files, still scanning")

            filelist_ondisk.append(filename_r.as_posix())

    # notify each other rank that there are no more files left

    other_ranks = set([a for a in range(1, ncpus)])
    for n in range(1,ncpus):
        comm.recv(None, source=MPI.ANY_SOURCE, tag=REQUEST_FILENAME_TAG, status=status)
        destination=status.Get_source()
        logger.debug(f"Rank 0: sending None (= no more files to verify) to {destination}")
        comm.send(None, dest=destination, tag=GET_FILENAME_TAG)
        other_ranks.remove(destination)

    if len(other_ranks) > 0:
        logger.error(f"Ranks {other_ranks} were not notified of end of file scanning.")

    return nfiles_ondisk

def compare_archive_checksums(target_dir: str, archive_files: str, summary_file: str|None):

    comm=MPI.COMM_WORLD
    rank=comm.Get_rank()
    ncpus=comm.Get_size()

    if ncpus == 1:
        logger.error("The number of MPI processes (NUM_TASKS) must be at least 2. Aborting.")
        comm.Abort()

    # read the contents of each archive. make available to all ranks.
    archive_type = determine_format(archive_files)

    if rank==0:
        if archive_type == 'h5':
            filelist_inarch, numbered_archives, qc_passed = check_archive_files(archive_files, logger)
        elif archive_type == 'zip':
            filelist_inarch, numbered_archives, qc_passed = check_zip_files(archive_files, logger)
        if not qc_passed:
            logger.error("Archive files failed initial quality control")
            comm.Abort()

    else:
        filelist_inarch = {}
        numbered_archives = {}
     
    filelist_inarch = comm.bcast(filelist_inarch, root=0)
    numbered_archives = comm.bcast(numbered_archives, root=0)

    if rank==0:
        nfiles_ondisk=scan_and_distribute_files(target_dir, comm, ncpus)
        logger.debug(f"Rank 0: done scanning, found {nfiles_ondisk}")

        success = True
        filesize_ondisk = []
        filesize_inarch = []
        md5_ondisk = []
        md5_inarch = []
        file_passed = []
        number_of_chunks = []
        filelist_ondisk = []
        error_files = []

    else:
        success, file_passed, error_files, filesize_ondisk, filesize_inarch, md5_ondisk, md5_inarch, filelist_ondisk, number_of_chunks = \
            run_verification(numbered_archives, filelist_inarch, target_dir, archive_type, rank, comm)

    comm.Barrier()

    # gather information from different processes
    overall_sucess = comm.allreduce(success, op=MPI.MIN)
    error_files_all = comm.gather(error_files)

    if rank==0:
        if overall_sucess:
            logger.info(f"archive(s) {archive_files} passed.")
        else:
            logger.error(f"archive(s) {archive_files} FAILED!.  The following files had errors:")
            logger.error(error_files_all)

    if summary_file is not None:

        filesize_ondisk_all = comm.gather(filesize_ondisk, root=0)
        filesize_inarch_all = comm.gather(filesize_inarch, root=0)
        
        md5_ondisk_all = comm.gather(md5_ondisk, root=0)
        md5_inarch_all = comm.gather(md5_inarch, root=0)

        file_passed_all = comm.gather(file_passed, root=0)
        # get the file list in the same order as the verification parameters
        filenames_all = comm.gather(filelist_ondisk, root=0)

        number_of_chunks = comm.gather(number_of_chunks, root=0)

        if rank==0:
            summary_info=  pd.DataFrame({"size_orig": flatten_list(filesize_ondisk_all),
                        "size_arch": flatten_list(filesize_inarch_all),
                        "md5_orig": flatten_list(md5_ondisk_all),
                        "md5_arch": flatten_list(md5_inarch_all),
                        "chunks": flatten_list(number_of_chunks),
                        "passed": flatten_list(file_passed_all)
                        }, index=flatten_list(filenames_all))

            summary_info.index.name = "filename"
            summary_info.to_json(summary_file)

    comm.Barrier()

            
def main():

    parser = argparse.ArgumentParser(
                prog='hdf5vault_check',
                description='Parallel tool to verify contents of archive ')

    parser.add_argument('-d', '--directory', required=True, type=str, help='name of directory to verify archive content against')
    parser.add_argument('-f', '--files', nargs="+", required=True, type=str, help="HDF5 archive file(s)")
    parser.add_argument('-j', '--json', required=False, type=str, help='JSON summary file (default: None)')
    parser.add_argument('-e', '--debug', action="store_true", help="log debugging information")
    parser.add_argument('-q', '--quiet', action="store_true", help='minimal output')

    args=parser.parse_args()

    rank=MPI.COMM_WORLD.Get_rank()

    target_dir = Path(args.directory)
    archive_files = args.files
    summary_file = args.json

    if args.debug:
        logger.setLevel(logging.DEBUG)
    elif args.quiet:
        logger.setLevel(logging.ERROR)
    else:
        logger.setLevel(logging.INFO)

    if rank==0:
        logger.info(f"Checking directory {target_dir} against archvive file(s) {archive_files}.")

    compare_archive_checksums(target_dir, archive_files, summary_file)

if __name__ == '__main__':
    main()



#! /usr/bin/env python

# An MPI (parallel) tool for unpacking HDF5 and zip archives
import argparse
from mpi4py import MPI
from pathlib import Path
import h5py
import zipfile
import blosc2
import logging
import sys
from humanize import intcomma, naturalsize
from time import time
from importlib.metadata import version

from common import check_archive_files, check_zip_files, determine_format

logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout)


# Split up dict with file archive information among CPUs
def scatter_items(filelist, ncpus):
    filelist_scattered=[{} for n in range(ncpus)]

    for n, key in enumerate(filelist.keys()):
        # first item goes to CPU 0, second to CPU 1, etc.
        cpu_num = n % ncpus

        filelist_scattered[cpu_num][key] = filelist[key]

    return filelist_scattered

def get_unpack_directory(dstdir: Path, files: list, comm: MPI.Comm):
    if len(files) == 1:
        tdir=Path(files[0]).stem

    else:
        basenames=[Path(f).stem for f in files]
        basenames=["_".join(f.split("_")[:-1]) for f in basenames]

        if len(set(basenames)) != 1: 
            logger.error('the expected format for multipart archives is archive_0.h5, archive_1.h5, etc...')
            comm.Abort()

        tdir=basenames[0]

    updir=dstdir.joinpath(tdir)

    return updir

def restore_file(filename: str, dstdir: str, archive_info: list, ctx: dict, archive_type: str, comm: MPI.Comm):

    nchunks = len(archive_info)
    # create dictionary that maps chunk number to entry in list (they might not be in order)
    cnmap = {a['chunk'] if a['chunk'] is not None else 0: p for p,a in enumerate(archive_info)}

    destfile = Path(dstdir).joinpath(filename)
    if destfile.is_file():
        logger.error(f'{destfile} already exists, aborting unpack.')
        comm.Abort()

    parentdir = destfile.parent

    # create output directory if it doesn't exist
    if not parentdir.is_dir():
        parentdir.mkdir(parents=True, exist_ok=True)

    restored_size=0

    with open(destfile, 'wb') as fid:
        for chunknum in range(nchunks):
            cinfo = archive_info[cnmap[chunknum]]

            archnum = cinfo["archive_number"]
            dataset=cinfo["dataset"]
            logger.debug(f"Restoring chunk {chunknum} from dataset {dataset} in archive {archnum}")

            if archive_type == 'h5':
                compdata = ctx[archnum][dataset][:].tobytes()
            elif archive_type == 'zip':
                compdata = ctx[archnum].read(dataset)
            else:
                logger.error(f"Unknown format type {archive_type}")
                comm.Abort()

            if cinfo["compression"] is None:
                rawdata=compdata

            elif cinfo["compression"].lower() in ['blosc','blosc2']:
                rawdata=blosc2.decompress(compdata)
            else:
                logger.error(f"Unknown compression method {cinfo['compression']} for file {filename}, part {chunknum}")
                comm.Abort()

            fid.write(rawdata)
            restored_size+=len(rawdata)

    return restored_size
 
def unpack_files(fileinfo: dict, numbered_archives: list[str], archive_type: str, dstdir:str, comm: MPI.Comm, progress_steps:int=100):
    rank=comm.Get_rank()
    nitems = len(fileinfo)

    restored_size = 0
    restored_files = 0

    # each archive is opened only once
    ctx={}
    for n in numbered_archives.keys():
        if archive_type == 'h5':
            ctx[n] = h5py.File(numbered_archives[n])
        elif archive_type == 'zip':
            ctx[n] = zipfile.ZipFile(numbered_archives[n])
        else:
            print(archive_type)
            logger.error('Unknown archive type.')

    # unpack each file
    for n,key in enumerate(fileinfo.keys()):
        if rank==0:
            if n % progress_steps == 0:
                logger.info(f'Unpacking file {n} out of {nitems} on rank 0')
        restored_size += restore_file(key, dstdir, fileinfo[key], ctx, archive_type, comm)
        restored_files += 1

    for n in ctx.keys():
        ctx[n].close()

    return restored_size, restored_files
    

def unpack_archives(files: list, dstdir: Path, format: str = 'H5'):

    comm=MPI.COMM_WORLD
    rank=comm.Get_rank()
    ncpus=comm.Get_size()

    if rank==0:
        archive_type=determine_format(files)
        if not archive_type:
            logger.error('Only .h5 and .zip files are supported.')
            comm.Abort()

    nfiles=len(files)

    t0=time()

    updir=get_unpack_directory(dstdir, files, comm)
    __version__ = version("hdf5vault")

    if rank==0:
        logger.info(f"HDF5/ZIP Archive unpacker, version f{__version__}.")
        logger.info(f"{archive_type} format detected.")
        logger.info(f"Unpacking {nfiles} files:")
        for file in files:
            logger.info(f" - {file}")
        logger.info(f"Destination directory: {dstdir}")


        if archive_type == 'h5':
            filelist_inarch,numbered_archives,qc_passed=check_archive_files(files, logger)
        elif archive_type == 'zip':
            filelist_inarch,numbered_archives,qc_passed=check_zip_files(files, logger)

        if not qc_passed:
            logger.error('QC not passed, aborting.')
            comm.Abort()

    else:
        filelist_inarch={}
        numbered_archives={}
        archive_type = ''

    filelist_inarch=comm.bcast(filelist_inarch, root=0)
    numbered_archives=comm.bcast(numbered_archives, root=0)
    archive_type=comm.bcast(archive_type, root=0)

    if rank==0:
        filelist_scattered = scatter_items(filelist_inarch, ncpus)
    else:
        filelist_scattered = []

    myfilelist = comm.scatter(filelist_scattered, root=0)

    logger.debug(f"Rank {rank}: received info for {len(myfilelist)} files to unpack.")
    nbytes_restored,nfiles_restored = unpack_files(myfilelist, numbered_archives, archive_type, updir, comm)

    logger.debug(f"Global rank {rank} at end of loop.")

    global_count=comm.allreduce(nfiles_restored, op=MPI.SUM)
    global_size=comm.allreduce(nbytes_restored, op=MPI.SUM)

    comm.Barrier()
    t1=time()
    if rank==0:
        logger.info(f"{intcomma(global_count)} files ({naturalsize(global_size)}) restored from {nfiles} archives in {t1-t0} seconds.")

def main():
    parser = argparse.ArgumentParser(
                prog='hdf5vault_unpack',
                description='Parallel tool to unpack HDF5 and zip archives')

    parser.add_argument('-f', '--files', nargs="+", required=True, type=str, help="archive file(s)")
    parser.add_argument('-d', '--destdir', required=False, type=Path, default=Path().cwd(), help='destination directory (default: current working directory)')
    parser.add_argument('-e', '--debug', action="store_true", help="log debugging information")
    parser.add_argument('-q', '--quiet', action="store_true", help='minimal output')

    args=parser.parse_args()

    if args.debug:
        logger.setLevel(logging.DEBUG)
    elif args.quiet:
        logger.setLevel(logging.WARN)
    else:
        logger.setLevel(logging.INFO)

    unpack_archives(files=args.files, dstdir=args.destdir) 

if __name__ == "__main__": 
    main()


HDF5_ARCHIVE_INFO= """
This HDF5 file is (part of) a file archive created with HDF5Vault (https://pypi.org/project/hdf5vault/). 

In this archive format, HDF5 is used as a hierarchival container for storing the content of a number of archived files.  Every HDF5 dataset in this archive contains the contents of a file or file chunk.  The (raw or compressed) bytes in the file or chunk are interpreted as 1-D dimensional array of type byte.  The dataset name mirrors the file name, while group names reflect directories and subdirectories in the archived data structure.

If the file is stored uncompressed and unchunked, the dataset name reflects the file name exactly. If the file has been split into chunks, the chunk number is appended to the dataset name (e.g., `___part1`). If the contents of the file have been compressed, the suffix `.blosc2` is added the the dataset name. 

This HDF5 file contains several attributes with additonal information:

__Description__: This text
__HD5Vault_version__: HDF5Vault version number
__HD5Vault_chunksize__: The size of a file chunk. Files smaller than this size are stored unchunked.
__HDF5Vault_info__: A JSON string with information on the chunking, compression and archive number of each archived file.

This file may be a part of a multi-file HDF5 archive. In this case, the following two attributes are present in each archive file:
__HDF5Vault_archive_file_number__: number of current file
__HDF5Vault_number_of_archive_files__: total number of files that make up the archive

EXAMPLE 

The original archived directory contains these files:
    README.txt
    directory1/small_datafile.bin
    directory2/big_datafile.bin

The resulting HDF5 archive(s) contains the datasets
    README.txt (dataset, type bytes)
    directory1/ (group)
    direcory1/small_datafile.bin.blosc2 (dataset, type bytes)
    direcory2/big_datafile.bin___part1.blosc2 (dataset, type bytes)
    direcory2/big_datafile.bin___part2.blosc2 (dataset, type bytes)

The HDF5 attribute __HDF5Vault_info__ is a JSON-formatted dictionary with information on the archived files.  Each key represents the path of the archived file, and the value is a list of dictionaries with storage information for each chunk, including chunk number (null if unchunked), dataset name, archive number, and compression used.

In the above example, __HDF5VAULT_info__  contains:

{
    "README.txt": [
        {
        "chunk": null,
        "archive_number": 0,
        "dataset": "README.txt",
        "compression": null
        }
    ],
    "directory1/small_datafile.bin": [
        {
        "chunk": null,
        "archive_number": 1,
        "dataset": "directory1/small_datafile.bin.blosc2",
        "compression": "blosc2"
        }
    ],
    "directory2/big_datafile.bin": [
        {
        "chunk": 0,
        "archive_number": 0,
        "dataset": "directory2/big_datafile.bin____part0.blosc2",
        "compression": "blosc2"
        },
        {
        "chunk": 1,
        "archive_number": 1,
        "dataset": "directory2/big_datafile.bin____part1.blosc2",
        "compression": "blosc2"
        }
            ]
}

That is, the file README.txt is stored unchunked and uncompressed in archive number 0 (example_0.h5).  The file directory2/big_datafile.bin is stored in two chunks located in archive 0 and 1, respectively, and each chunk was compressed using BLOSC2.  In multipart HDF5 archives, the attribute HDF5Vault is replicated in each archive file.

ARCHIVE UNPACKING

HDF5Vault comes with a parallel unpacking tool, hdf5vault_unpack. The following Python code snippet illustrates how the content of a compressed file in the archive can be restored:

    import h5py
    import blosc2

    with h5py.File('example_1.h5') as hfile:
        compressed_data = hfile["directory1/small_datafile.bin.blosc2"][:].tobytes()

    raw_data = blosc2.decompress(raw_data)

    with open("directory1/small_datafile.bin", "rb") as fid:
        fid.write(raw_data)

"""